@@ -9,33 +9,21 @@
 static const char hex_chars_lc[] = "0123456789abcdef";
 static const char hex_chars_uc[] = "0123456789ABCDEF";
 
-/**
- * init the buffer
- *
- */
 
 __attribute_noinline__
 buffer* buffer_init(void) {
+  #if 0 /* buffer_init() and chunk_init() can be hot,
+	 * so avoid the additional hop of indirection */
+	return ck_calloc(1, sizeof(buffer));
+  #else
 	buffer * const b = calloc(1, sizeof(*b));
 	force_assert(b);
 	return b;
-}
-
-buffer *buffer_init_buffer(const buffer *src) {
-	buffer * const b = buffer_init();
-	buffer_copy_string_len(b, BUF_PTR_LEN(src));
-	return b;
-}
-
-buffer *buffer_init_string(const char *str) {
-	buffer * const b = buffer_init();
-	buffer_copy_string(b, str);
-	return b;
+  #endif
 }
 
 void buffer_free(buffer *b) {
 	if (NULL == b) return;
-
 	free(b->ptr);
 	free(b);
 }
@@ -170,8 +158,10 @@ void buffer_commit(buffer *b, size_t size)
 	b->ptr[sz - 1] = '\0';
 }
 
+__attribute_cold__ /*(reduce code size due to inlining)*/
 void buffer_copy_string(buffer * restrict b, const char * restrict s) {
-	buffer_copy_string_len(b, s, NULL != s ? strlen(s) : 0);
+    if (__builtin_expect( (NULL == s), 0)) s = "";
+    buffer_copy_string_len(b, s, strlen(s));
 }
 
 void buffer_copy_string_len(buffer * const restrict b, const char * const restrict s, const size_t len) {
@@ -183,8 +173,10 @@ void buffer_copy_string_len(buffer * const restrict b, const char * const restri
     memcpy(d, s, len);
 }
 
+__attribute_cold__ /*(reduce code size due to inlining)*/
 void buffer_append_string(buffer * restrict b, const char * restrict s) {
-	buffer_append_string_len(b, s, NULL != s ? strlen(s) : 0);
+    if (__builtin_expect( (NULL == s), 0)) s = "";
+    buffer_append_string_len(b, s, strlen(s));
 }
 
 /**
@@ -241,8 +233,14 @@ void buffer_append_iovec(buffer * const restrict b, const struct const_iovec * c
 
 void buffer_append_path_len(buffer * restrict b, const char * restrict a, size_t alen) {
     char * restrict s = buffer_string_prepare_append(b, alen+1);
+  #ifdef _WIN32
+    const int aslash = (alen && (a[0] == '/' || a[0] == '\\'));
+    if (b->used > 1 && (s[-1] == '/' || s[-1] == '\\'))
+  #else
     const int aslash = (alen && a[0] == '/');
-    if (b->used > 1 && s[-1] == '/') {
+    if (b->used > 1 && s[-1] == '/')
+  #endif
+    {
         if (aslash) {
             ++a;
             --alen;
@@ -304,12 +302,10 @@ __attribute_nonnull__()
 __attribute_returns_nonnull__
 static char* utostr(char buf[LI_ITOSTRING_LENGTH], uintmax_t val) {
 	char *cur = buf+LI_ITOSTRING_LENGTH;
+	uintmax_t x;
 	do {
-		int mod = val % 10;
-		val /= 10;
-		/* prepend digit mod */
-		*(--cur) = (char) ('0' + mod);
-	} while (0 != val);
+		*(--cur) = (char) ('0' + (int)(val - (x = val/10) * 10));
+	} while (0 != (val = x));           /* val % 10 */
 	return cur;
 }
 
@@ -334,7 +330,6 @@ void buffer_append_int(buffer *b, intmax_t val) {
 }
 
 void buffer_append_strftime(buffer * const restrict b, const char * const restrict format, const struct tm * const restrict tm) {
-    force_assert(NULL != format);
     /*(localtime_r() or gmtime_r() producing tm should not have failed)*/
     if (__builtin_expect( (NULL == tm), 0)) return;
 
@@ -413,13 +408,7 @@ int buffer_eq_icase_ssn(const char * const a, const char * const b, const size_t
     for (size_t i = 0; i < len; ++i) {
         unsigned int ca = ((unsigned char *)a)[i];
         unsigned int cb = ((unsigned char *)b)[i];
-        if (ca != cb) {
-            ca |= 0x20;
-            cb |= 0x20;
-            if (ca != cb) return 0;
-            if (!light_islower(ca)) return 0;
-            if (!light_islower(cb)) return 0;
-        }
+        if (ca != cb && ((ca ^ cb) != 0x20 || !light_isalpha(ca))) return 0;
     }
     return 1;
 }
@@ -482,7 +471,7 @@ void buffer_substr_replace (buffer * const restrict b, const size_t offset,
     const size_t rlen = buffer_clen(replace);
 
     if (rlen > len) {
-        buffer_extend(b, blen-len+rlen);
+        buffer_extend(b, rlen-len);
         memmove(b->ptr+offset+rlen, b->ptr+offset+len, blen-offset-len);
     }
 
@@ -609,10 +598,6 @@ void buffer_append_string_encoded(buffer * const restrict b, const char * const
 	size_t d_len, ndx;
 	const char *map = NULL;
 
-	if (0 == s_len) return;
-
-	force_assert(NULL != s);
-
 	switch(encoding) {
 	case ENCODING_REL_URI:
 		map = encoded_chars_rel_uri;
@@ -628,8 +613,6 @@ void buffer_append_string_encoded(buffer * const restrict b, const char * const
 		break;
 	}
 
-	force_assert(NULL != map);
-
 	/* count to-be-encoded-characters */
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
 		if (map[*ds & 0xFF]) {
@@ -684,12 +667,11 @@ void buffer_append_string_c_escaped(buffer * const restrict b, const char * cons
 	unsigned char *ds, *d;
 	size_t d_len, ndx;
 
-	if (0 == s_len) return;
-
 	/* count to-be-encoded-characters */
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
-		if ((*ds < 0x20) /* control character */
-				|| (*ds >= 0x7f)) { /* DEL + non-ASCII characters */
+		if (__builtin_expect( (*ds >= ' ' && *ds <= '~'), 1))
+			d_len++;
+		else { /* CTLs or non-ASCII characters */
 			switch (*ds) {
 			case '\t':
 			case '\r':
@@ -700,8 +682,6 @@ void buffer_append_string_c_escaped(buffer * const restrict b, const char * cons
 				d_len += 4; /* \xCC */
 				break;
 			}
-		} else {
-			d_len++;
 		}
 	}
 
@@ -713,8 +693,9 @@ void buffer_append_string_c_escaped(buffer * const restrict b, const char * cons
 	}
 
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
-		if ((*ds < 0x20) /* control character */
-				|| (*ds >= 0x7f)) { /* DEL + non-ASCII characters */
+		if (__builtin_expect( (*ds >= ' ' && *ds <= '~'), 1))
+			d[d_len++] = *ds;
+		else { /* CTLs or non-ASCII characters */
 			d[d_len++] = '\\';
 			switch (*ds) {
 			case '\t':
@@ -728,17 +709,116 @@ void buffer_append_string_c_escaped(buffer * const restrict b, const char * cons
 				break;
 			default:
 				d[d_len++] = 'x';
-				d[d_len++] = hex_chars_lc[((*ds) >> 4) & 0x0F];
+				d[d_len++] = hex_chars_lc[(*ds) >> 4];
 				d[d_len++] = hex_chars_lc[(*ds) & 0x0F];
 				break;
 			}
-		} else {
-			d[d_len++] = *ds;
 		}
 	}
 }
 
 
+void
+buffer_append_bs_escaped (buffer * const restrict b,
+                          const char * restrict s, const size_t len)
+{
+    /* replaces non-printable chars with escaped string
+     * default: \xHH where HH is the hex representation of the byte
+     * exceptions: " => \", \ => \\, whitespace chars => \n \t etc. */
+    /* Intended for use escaping string to be surrounded by double-quotes */
+    /* Performs single pass over string and is optimized for ASCII;
+     * non-ASCII escaping might be slightly sped up by walking input twice,
+     * first to calculate escaped length and extend the destination b, and
+     * second to do the escaping. (This non-ASCII optim is not done here) */
+    buffer_string_prepare_append(b, len);
+    for (const char * const end = s+len; s < end; ++s) {
+        unsigned int c;
+        const char * const ptr = s;
+        do {
+            c = *(const unsigned char *)s;
+        } while (c >= ' ' && c <= '~' && c != '"' && c != '\\' && ++s < end);
+        if (s - ptr) buffer_append_string_len(b, ptr, s - ptr);
+
+        if (s == end)
+            return;
+
+        /* ('\a', '\v' shortcuts are technically not json-escaping) */
+        /* ('\0' is also omitted due to the possibility of string corruption if
+         *  the receiver supports decoding octal escapes (\000) and the escaped
+         *  string contains \0 followed by two digits not part of escaping)*/
+
+        char *d;
+        switch (c) {
+          case '\a':case '\b':case '\t':case '\n':case '\v':case '\f':case '\r':
+            c = "0000000abtnvfr"[c];
+            __attribute_fallthrough__
+          case '"': case '\\':
+            d = buffer_extend(b, 2);
+            d[0] = '\\';
+            d[1] = c;
+            break;
+          default:
+            /* non printable char => \xHH */
+            d = buffer_extend(b, 4);
+            d[0] = '\\';
+            d[1] = 'x';
+            d[2] = hex_chars_uc[c >> 4];
+            d[3] = hex_chars_uc[c & 0xF];
+            break;
+        }
+    }
+}
+
+
+void
+buffer_append_bs_escaped_json (buffer * const restrict b,
+                               const char * restrict s, const size_t len)
+{
+    /* replaces non-printable chars with escaped string
+     * json: \u00HH where HH is the hex representation of the byte
+     * exceptions: " => \", \ => \\, whitespace chars => \n \t etc. */
+    /* Intended for use escaping string to be surrounded by double-quotes */
+    buffer_string_prepare_append(b, len);
+    for (const char * const end = s+len; s < end; ++s) {
+        unsigned int c;
+        const char * const ptr = s;
+        do {
+            c = *(const unsigned char *)s;
+        } while (c >= ' ' && c != '"' && c != '\\' && ++s < end);
+        if (s - ptr) buffer_append_string_len(b, ptr, s - ptr);
+
+        if (s == end)
+            return;
+
+        /* ('\a', '\v' shortcuts are technically not json-escaping) */
+        /* ('\0' is also omitted due to the possibility of string corruption if
+         *  the receiver supports decoding octal escapes (\000) and the escaped
+         *  string contains \0 followed by two digits not part of escaping)*/
+
+        char *d;
+        switch (c) {
+          case '\a':case '\b':case '\t':case '\n':case '\v':case '\f':case '\r':
+            c = "0000000abtnvfr"[c];
+            __attribute_fallthrough__
+          case '"': case '\\':
+            d = buffer_extend(b, 2);
+            d[0] = '\\';
+            d[1] = c;
+            break;
+          default:
+            d = buffer_extend(b, 6);
+            d[0] = '\\';
+            d[1] = 'u';
+            d[2] = '0';
+            d[3] = '0';
+            d[4] = hex_chars_uc[c >> 4];
+            d[5] = hex_chars_uc[c & 0xF];
+            break;
+        }
+    }
+}
+
+
 /* decodes url-special-chars inplace.
  * replaces non-printable characters with '_'
  * (If this is used on a portion of query string, then query string should be
@@ -768,6 +848,7 @@ void buffer_urldecode_path(buffer * const b) {
 
 int buffer_is_valid_UTF8(const buffer *b) {
     /* https://www.w3.org/International/questions/qa-forms-utf-8 */
+    /*assert(b->used);*//*(b->ptr must exist and be '\0'-terminated)*/
     const unsigned char *c = (unsigned char *)b->ptr;
     while (*c) {
 
@@ -825,7 +906,7 @@ void buffer_path_simplify(buffer *b)
         return;
     }
 
-  #if defined(__WIN32) || defined(__CYGWIN__)
+  #if defined(_WIN32) || defined(__CYGWIN__)
     /* cygwin is treating \ and / the same, so we have to that too */
     for (char *p = b->ptr; *p; p++) {
         if (*p == '\\') *p = '/';
@@ -910,7 +991,8 @@ void buffer_path_simplify(buffer *b)
 
 void buffer_to_lower(buffer * const b) {
     unsigned char * const restrict s = (unsigned char *)b->ptr;
-    for (uint32_t i = 0; i < b->used; ++i) {
+    const uint_fast32_t used = b->used;
+    for (uint_fast32_t i = 0; i < used; ++i) {
         if (light_isupper(s[i])) s[i] |= 0x20;
     }
 }
@@ -918,7 +1000,8 @@ void buffer_to_lower(buffer * const b) {
 
 void buffer_to_upper(buffer * const b) {
     unsigned char * const restrict s = (unsigned char *)b->ptr;
-    for (uint32_t i = 0; i < b->used; ++i) {
+    const uint_fast32_t used = b->used;
+    for (uint_fast32_t i = 0; i < used; ++i) {
         if (light_islower(s[i])) s[i] &= 0xdf;
     }
 }
